/* vim: set ai et ts=4 sw=4: */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h>
#include <gnome.h>
#include <id3tag.h>
#include <vorbis/codec.h>
#include <vorbis/vorbisfile.h>

#include "tag.h"
#include "util.h"
#include "decoder.h"
#include "prefs.h"

/*-----------------------------------------------------------------*/

static GSList *fallback_codesets = NULL;
static gchar *used_codeset = NULL;

typedef struct _TagHandler {
    gboolean (*tag_read_func)(Tag* tag);
    gboolean (*tag_write_func)(Tag* tag);
    gboolean (*tag_remove_func)(Tag* tag);
} TagHandler;


static gboolean tag_read_id3(Tag* tag);
static gboolean tag_write_id3(Tag* tag);
static gboolean tag_remove_id3(Tag* tag);

static gboolean tag_read_ogg(Tag* tag);
static gboolean tag_write_ogg(Tag* tag);
static gboolean tag_remove_ogg(Tag* tag);

static gboolean tag_read_unknown(Tag* tag);
static gboolean tag_write_unknown(Tag* tag);
static gboolean tag_remove_unknown(Tag* tag);


TagHandler tag_handlers[] = {
    { tag_read_id3, tag_write_id3, tag_remove_id3 },
    { tag_read_ogg, tag_write_ogg, tag_remove_ogg },
    { tag_read_unknown, tag_write_unknown, tag_remove_unknown },
};

/*-----------------------------------------------------------------*/

//
// common supports
//

//@@backport{{
void tag_init(void)
{
	gint index;
	gchar **array;
	g_slist_free(fallback_codesets);
	fallback_codesets = NULL;
	array = g_strsplit(prefs.tag_codesets, ",", 0);
	for(index = 0; array[index] != NULL; index++) {
	    fallback_codesets = g_slist_append(fallback_codesets, array[index]);
	}
}
//@@}}

Tag* tag_new(void)
{
    Tag* tag;
    tag = g_new0(Tag, 1);
	tag->type = TAG_TYPE_UNKNOWN;
    return tag;
}


Tag* tag_new_from_file(const gchar* filename)
{
    Tag* tag;

    tag = tag_new();
    tag->filename = g_strdup(filename);
    tag->type = tag_get_type(filename);
    if(!tag_read(tag)) {
        tag->type = TAG_TYPE_UNKNOWN;
        tag_read(tag); // this returns true always ;)
    }

    return tag;
}


void tag_free(Tag* tag)
{
    g_free(tag->filename);
    g_free(tag->title);
    g_free(tag->artist);
    g_free(tag->album);
    g_free(tag->genre);
    g_free(tag->comment);
    g_free(tag);
}


/** read tag from file */
gboolean tag_read(Tag* tag)
{
	gchar *temp_str;
	gboolean ret;
	
	ret = tag_handlers[tag->type].tag_read_func(tag);

//@@backport{{
	// provides fallback title from uri
	if (LA_STR_IS_EMPTY(tag->title)) {
		temp_str = g_path_get_basename(tag->filename);
		tag->title = filename_to_utf8(temp_str);
		g_free(temp_str);
	}
//@@}}

	return ret;
}

/** write tag to file */
gboolean tag_write(Tag* tag)
{
    return tag_handlers[tag->type].tag_write_func(tag);
}


/** remove tag from file */
gboolean tag_remove(Tag* tag)
{
    return tag_handlers[tag->type].tag_remove_func(tag);
}


/** guess tag type from file extension */
TagType tag_get_type(const gchar* filename)
{
	if(str_has_suffix_nocase(filename, ".ogg"))
        return TAG_TYPE_OGG;
	if(str_has_suffix_nocase(filename, ".mp3"))
        return TAG_TYPE_ID3V1;
    return TAG_TYPE_UNKNOWN;
}

/*-----------------------------------------------------------------*/

//
// unknown tag support(matches with all but recognized tag)
//

static gboolean tag_read_unknown(Tag* tag)
{
    gchar* filename = filename_to_utf8(tag->filename);
    tag->title = g_path_get_basename(filename);
    g_free(filename);
    return TRUE;
}

static gboolean tag_write_unknown(Tag* tag)
{
    g_print("tag_write_unknown:%s\n", tag->filename);
    return TRUE;
}

static gboolean tag_remove_unknown(Tag* tag)
{
    g_print("tag_remove_unknown:%s\n", tag->filename);
    return TRUE;
}

/*-----------------------------------------------------------------*/

static gchar *get_id3_tag_string(struct id3_tag *tag, char *id)
{
	struct id3_frame *frame;
	gboolean is_genre_frame;
	GString *str;
	union id3_field *field;
	gint nstrings, index;
	id3_ucs4_t const *ucs4;
	gchar *latin1;
	gchar *utf8;

	frame = id3_tag_findframe(tag, id, 0);
	if (frame) {
		is_genre_frame = !strcasecmp(id, ID3_FRAME_GENRE);

		str = g_string_sized_new(32);

		field = &frame->fields[1];
		nstrings = id3_field_getnstrings(field);

		for (index = 0; index < nstrings; index++) {
			ucs4 = id3_field_getstrings(field, index);
			if (!ucs4)
				return NULL;

			if (is_genre_frame)
				ucs4 = id3_genre_name(ucs4);

			// some troubles with utf8duplicate :(
			// so, i'd used latin1 and convert it utf8 by hand
			//utf8 = id3_ucs4_utf8duplicate(ucs4);
			//if (!utf8) break;
			//g_string_append(str, utf8);

			latin1 = id3_ucs4_latin1duplicate(ucs4);
			if (!latin1)
				break;
			g_string_append(str, latin1);
		}

		//utf8 = str->str;
		//g_string_free(str, FALSE);
		//return utf8;

		utf8 =
		    la_str_to_utf8_with_fallback(str->str,
						 fallback_codesets,
						 &used_codeset);
		g_string_free(str, FALSE);
		return utf8;
	}

	return NULL;
}

static gchar *get_id3_tag_comment(struct id3_tag *tag)
{
	struct id3_frame *frame;
	id3_ucs4_t const *ucs4;
	gint index;
	gchar *latin1;
	gchar *utf8;

	index = 0;
	while ((frame =
		id3_tag_findframe(tag, ID3_FRAME_COMMENT, index++))) {
		id3_latin1_t *ptr, *newline;
		int first = 1;

		ucs4 = id3_field_getstring(&frame->fields[2]);
		if (!ucs4)
			return NULL;

		if (*ucs4)
			continue;

		ucs4 = id3_field_getfullstring(&frame->fields[3]);
		if (!ucs4)
			return NULL;

		latin1 = id3_ucs4_latin1duplicate(ucs4);

		utf8 =
		    la_str_to_utf8_with_fallback(latin1, fallback_codesets,
						 &used_codeset);
		g_free(latin1);
		return utf8;
	}

	return NULL;
}

static gboolean tag_read_id3(Tag* tag)
{
	struct id3_file *id3_file;
	struct id3_tag *id3_tag;

	// sample properties
	tag->channels = 2;
	tag->rate = 0;
	tag->bitrate = 0;

	if((decoder_get_total_time_frame(tag->filename,&(tag->seconds),&(tag->frames))) == -1) {
		tag->seconds = 0;
		tag->frames = 0;
		g_print("Invalid file %s\n",tag->filename);
		return FALSE;
	};

	if(!prefs.parse_tag) {
		return FALSE;
	}

	id3_file = id3_file_open(tag->filename, ID3_FILE_MODE_READONLY);
	if (!id3_file)
		return FALSE;

	id3_tag = id3_file_tag(id3_file);
	if (!id3_tag)
		return FALSE;

	tag->title = get_id3_tag_string(id3_tag, ID3_FRAME_TITLE);
	tag->artist= get_id3_tag_string(id3_tag, ID3_FRAME_ARTIST);
	tag->album = get_id3_tag_string(id3_tag, ID3_FRAME_ALBUM);
	tag->track = la_strtol(get_id3_tag_string(id3_tag, ID3_FRAME_TRACK), -1);
	tag->year = la_strtol(get_id3_tag_string(id3_tag, ID3_FRAME_YEAR), -1);
	tag->genre = get_id3_tag_string(id3_tag, ID3_FRAME_GENRE);
	tag->comment = get_id3_tag_comment(id3_tag);

	id3_tag_delete(id3_tag);
	id3_file_close(id3_file);

	return TRUE;
}

static gboolean tag_write_id3(Tag* tag)
{
    g_print("NYI tag_write_id3:%s\n", tag->filename);
    return TRUE;
}


static gboolean tag_remove_id3(Tag* tag)
{
    g_print("NYI tag_remove_id3:%s\n", tag->filename);
    return TRUE;
}

/*-----------------------------------------------------------------*/

//
// ogg-vorbis supports
//

static gboolean tag_read_ogg(Tag* tag)
{
	FILE* fp;
    gchar* temp_str;
    OggVorbis_File vf;
    vorbis_info* vi;
	GString *comment_str;
	char **ptr;
	gchar **tokens;
	gchar *utf8;

	if((fp = fopen(tag->filename, "rb")) == NULL) {
		g_print("Can't open %s\n", tag->filename);
		return FALSE;
	}

    if(ov_open(fp, &vf, NULL, 0) < 0) {
        g_print("%s does not appear to be an Ogg bitstream.\n", tag->filename);
        return FALSE;
    }

    // sample properties
    
    vi = ov_info(&vf, -1);
    tag->channels = vi->channels;
    tag->rate = vi->rate;
    tag->bitrate = vi->bitrate_nominal;
    tag->frames = ov_pcm_total(&vf, -1);
    tag->seconds = ov_time_total(&vf, -1);

	if(!prefs.parse_tag) {
		ov_clear(&vf);
		return FALSE;
	}

	comment_str = g_string_sized_new(32);

	ptr = ov_comment(&vf, -1)->user_comments;
	while (*ptr) {
		tokens = g_strsplit(*ptr, "=", 2);

		if (LA_STR_IS_NOT_EMPTY(tokens[1])) {
			utf8 =
			    la_str_to_utf8_with_fallback(tokens[1],
							 fallback_codesets,
							 &used_codeset);

			if(!strcasecmp(tokens[0], "title"))
				tag->title = g_strdup(utf8);
			else if(!strcasecmp(tokens[0], "artist"))
				tag->artist = g_strdup(utf8);
			else if(!strcasecmp(tokens[0], "album"))
				tag->album = g_strdup(utf8);
			else if(!strcasecmp(tokens[0], "track") || !strcasecmp(tokens[0], "tracknumber"))
				tag->track = la_strtol(utf8, -1);
			else if(!strcasecmp(tokens[0], "year"))
				tag->year = la_strtol(utf8, -1);
			else if(!strcasecmp(tokens[0], "genre"))
				tag->genre = g_strdup(utf8);

			g_free(utf8);
		} else {
			g_string_append(comment_str, *ptr);
			g_string_append_c(comment_str, '\n');
		}
		g_strfreev(tokens);

		++ptr;
	}

	ov_clear(&vf);

	if (LA_STR_IS_NOT_EMPTY(comment_str->str)) {
		utf8 =
		    la_str_to_utf8_with_fallback(comment_str->str,
						 fallback_codesets,
						 &used_codeset);
		tag->comment = utf8;
	}

	g_string_free(comment_str, TRUE);

	return TRUE;
}


static gboolean tag_write_ogg(Tag* tag)
{
    g_print("NYI tag_write_ogg:%s\n", tag->filename);
    return TRUE;
}


static gboolean tag_remove_ogg(Tag* tag)
{
    g_print("NYI tag_remove_ogg:%s\n", tag->filename);
    return TRUE;
}

/*tag.c*/
